from collections import OrderedDict

import tensorflow as tf
from tensorflow.python.keras.layers import (Embedding, Input, Concatenate, Flatten, Dense)
from tensorflow.python.keras.regularizers import l2

from dup1.feature.feature_column import SparseFeat, DenseFeat, VarLenSparseFeat
from dup1.feature.feature_column import build_input_features
from dup1.feature.inputs import embedding_lookup, get_dense_input, get_embedding_vec_list
from dup1.layers.bilstm import BiLSTM
from dup1.layers.core import DNN, PredictionLayer
from dup1.layers.encoding import BiasEncoding
from dup1.layers.sequence import AttentionSequencePoolingLayer
from dup1.layers.transformer import Transformer
from dup1.layers.utils import concat_func


def sess_interest_division(sparse_embedding_dict, user_behavior_input_dict, sparse_fg_list, sess_feture_list,
                           sess_max_count,
                           bias_encoding=True):
    tr_input = []
    for i in range(sess_max_count):
        sess_name = "sess_" + str(i)
        keys_emb_list = get_embedding_vec_list(sparse_embedding_dict, user_behavior_input_dict[sess_name],
                                               sparse_fg_list, sess_feture_list, sess_feture_list)

        keys_emb = concat_func(keys_emb_list, mask=True)
        tr_input.append(keys_emb)
    if bias_encoding:
        tr_input = BiasEncoding(sess_max_count)(tr_input)
    return tr_input


def sess_interest_extractor(tr_input, sess_max_count, TR):
    tr_out = []
    for i in range(sess_max_count):
        tr_out.append(TR(
            [tr_input[i], tr_input[i]]))
    sess_fea = concat_func(tr_out, axis=1)
    return sess_fea


class DSIN:
    def __init__(self, dnn_feature_columns, sess_feature_list, sess_max_count=5, bias_encoding=False,
                 att_embedding_size=1, att_head_num=8, dnn_hidden_units=(200, 80), dnn_activation='sigmoid',
                 dnn_dropout=0,
                 dnn_use_bn=False, l2_reg_dnn=0, l2_reg_embedding=1e-6, seed=1024, task='binary'):
        hist_emb_size = sum(
            map(lambda fc: fc.embedding_dim, filter(lambda fc: fc.name in sess_feature_list, dnn_feature_columns)))
        print(hist_emb_size)

        features = build_input_features(dnn_feature_columns)
        print(features)

        sparse_feature_columns = list(
            filter(lambda x: isinstance(x, SparseFeat), dnn_feature_columns)) if dnn_feature_columns else []
        dense_feature_columns = list(
            filter(lambda x: isinstance(x, DenseFeat), dnn_feature_columns)) if dnn_feature_columns else []
        varlen_sparse_feature_columns = list(
            filter(lambda x: isinstance(x, VarLenSparseFeat), dnn_feature_columns)) if dnn_feature_columns else []

        print(sparse_feature_columns)
        print(dense_feature_columns)
        print(varlen_sparse_feature_columns)

        sparse_varlen_feature_columns = []
        history_fc_names = list(map(lambda x: "sess" + x, sess_feature_list))
        for fc in varlen_sparse_feature_columns:
            feature_name = fc.name
            if feature_name in history_fc_names:
                continue
            else:
                sparse_varlen_feature_columns.append(fc)

        inputs_list = list(features.values())

        user_behavior_input_dict = {}
        for idx in range(sess_max_count):
            sess_input = OrderedDict()
            for i, feat in enumerate(sess_feature_list):
                sess_input[feat] = features["sess_" + str(idx) + "_" + feat]
            user_behavior_input_dict["sess_" + str(idx)] = sess_input

        user_sess_length = Input(shape=(1,), name="sess_length_ph")
        print(inputs_list, user_sess_length)

        embedding_dict = {feat.embedding_name: Embedding(feat.vocabulary_size, feat.embedding_dim,
                                                         embeddings_initializer=feat.embeddings_initializer,
                                                         embeddings_regularizer=l2(
                                                             l2_reg_embedding),
                                                         name='sparse_emb_' +
                                                              str(i) + '-' + feat.name,
                                                         mask_zero=(feat.name in sess_feature_list)) for i, feat in
                          enumerate(sparse_feature_columns)}
        print(embedding_dict)

        label_ph = Input(shape=(1,), name='label_ph')
        # input_ph_list = inputs_list + [user_sess_length]
        features["sess_length_ph"] = user_sess_length
        features["label_ph"] = label_ph
        self.input_ph_dict = features

        query_emb_list = embedding_lookup(embedding_dict, features, sparse_feature_columns, sess_feature_list,
                                          sess_feature_list, to_list=True)
        dnn_input_emb_list = embedding_lookup(embedding_dict, features, sparse_feature_columns,
                                              mask_feat_list=sess_feature_list, to_list=True)
        dense_value_list = get_dense_input(features, dense_feature_columns)
        query_emb = concat_func(query_emb_list, mask=True)

        print(query_emb, dnn_input_emb_list, dense_value_list)
        dnn_input_emb = Flatten()(concat_func(dnn_input_emb_list))
        if len(dense_value_list) > 0:
            dense_input_emb = concat_func(dense_value_list)
            dnn_input_emb = concat_func([dnn_input_emb, dense_input_emb])
        print(dnn_input_emb)

        tr_input = sess_interest_division(embedding_dict, user_behavior_input_dict, sparse_feature_columns,
                                          sess_feature_list, sess_max_count, bias_encoding=bias_encoding)
        print(tr_input)
        Self_Attention = Transformer(att_embedding_size, att_head_num, dropout_rate=0, use_layer_norm=False,
                                     use_positional_encoding=(not bias_encoding), seed=seed, supports_masking=True,
                                     blinding=True)
        print(Self_Attention)
        sess_fea = sess_interest_extractor(
            tr_input, sess_max_count, Self_Attention)

        interest_attention_layer = AttentionSequencePoolingLayer(att_hidden_units=(64, 16), weight_normalization=True,
                                                                 supports_masking=False)(
            [query_emb, sess_fea, user_sess_length])
        print(interest_attention_layer)

        lstm_outputs = BiLSTM(hist_emb_size,
                              layers=2, res_layers=0, dropout_rate=0.2, )(sess_fea)
        lstm_attention_layer = AttentionSequencePoolingLayer(att_hidden_units=(64, 16), weight_normalization=True)(
            [query_emb, lstm_outputs, user_sess_length])
        print(lstm_attention_layer)

        dnn_input_emb = Concatenate()(
            [dnn_input_emb, Flatten()(interest_attention_layer), Flatten()(lstm_attention_layer)])

        output = DNN(dnn_hidden_units, dnn_activation, l2_reg_dnn, dnn_dropout, dnn_use_bn, seed=seed)(dnn_input_emb)
        output = Dense(1, use_bias=False, kernel_initializer=tf.keras.initializers.glorot_normal(seed))(output)
        output = PredictionLayer(task)(output)
        score = tf.identity(output, name="score")
        self.score = score
        self.end_point = {"logits": score}
        self.build_loss(score, label_ph)
        self.build_optim()

    def build_loss(self, logits, labels, **kwargs):
        """
        :param logits:
        :param labels:
        :param kwargs:
        :return:
        """
        labels = tf.cast(labels, tf.float32)
        self.loss = tf.reduce_mean(tf.keras.losses.binary_crossentropy(y_true=labels, y_pred=logits))


    def build_optim(self, **kwargs):
        """
        :param kwargs:
        :return:
        """
        self.optim = tf.train.AdamOptimizer(learning_rate=0.001, beta1=0.9, beta2=0.999, epsilon=1e-08)

