import copy
import json
import math
import os
import random
import time

import numpy as np
import tensorflow as tf

from dup1.tasks.task_utils import trans_feat_feed
from dup1.utils.com_util import list_files
from dup1.utils.log_util import pylog
from tfconf import log_dir, global_parameters


def build_train_op(issync, optim_op, loss_op, tf_json):
    """
    :param issync:
    :param optim_op:
    :param loss_op:
    :param tf_json:
    :return:
    """
    init_token_op, chief_queue_runner = None, None
    step_op = tf.train.get_or_create_global_step()
    grads_and_vars = optim_op.compute_gradients(loss_op)
    if issync == 1:
        # 同步模式计算更新梯度
        rep_op = tf.train.SyncReplicasOptimizer(optim_op,
                                                replicas_to_aggregate=len(tf_json.get("cluster").get("worker")),
                                                # replica_id=FLAGS.task_index,
                                                total_num_replicas=len(tf_json.get("cluster").get("worker")),
                                                use_locking=True)
        train_op = rep_op.apply_gradients(grads_and_vars, global_step=step_op)
        init_token_op = rep_op.get_init_tokens_op()
        chief_queue_runner = rep_op.get_chief_queue_runner()
    else:
        # 异步模式计算更新梯度
        train_op = optim_op.apply_gradients(grads_and_vars, global_step=step_op)
    return train_op, step_op, loss_op, init_token_op, chief_queue_runner


def build_work_readers(reader, tf_json):
    """
    :param reader:
    :param tf_json:
    :return:
    """
    worker_reader = []
    worker_count = int(tf_json["worker"]["count"])
    worker_list = [[] for _ in range(worker_count)]
    for path in reader.filenames:
        files = list_files(path)
        pylog.info(f"[Trainer] Total Train file: {len(files)} Train file 0: {files[0]}")
        random.Random(1).shuffle(files)
        group_count = math.ceil(len(files) / worker_count)
        if group_count > 1:
            worker_files = [files[i * group_count: (i + 1) * group_count] for i in range(worker_count)]
        else:
            worker_files = [files for _ in range(worker_count)]
        for i in range(worker_count):
            worker_list[i].extend(worker_files[i])
    for work_files in worker_list:
        reader.filenames = work_files
        worker_reader.append(copy.copy(reader))
    return worker_reader


class DistPSTask:
    def __init__(self, reader, model, batch_size=256, steps_log=32, steps_save=1024):
        self.reader = reader
        self.model = model
        self.batch_size = batch_size
        self.steps_log = steps_log
        self.steps_save = steps_save

    def train(self):
        cluster_def = eval(os.getenv("TF_CONFIG"))["cluster"]
        cluster = tf.train.ClusterSpec(cluster_def)
        tf_json = json.loads(os.environ['TF_CONFIG'])
        flag_job_name = tf_json.get("task").get("type")
        flag_task_index = tf_json.get("task").get("index")
        server = tf.train.Server(cluster, job_name=flag_job_name, task_index=flag_task_index)
        issync = global_parameters["issync"]
        readers = build_work_readers(self.reader, tf_json)
        if flag_job_name == "ps":
            server.join()
        elif flag_job_name == "worker":
            with tf.device(tf.train.replica_device_setter(
                    worker_device="/job:worker/task:%d" % flag_task_index, cluster=cluster)):
                train_op, step_op, loss_op, init_token_op, chief_queue_runner = build_train_op(issync,
                                                                                               self.model.optim,
                                                                                               self.model.loss,
                                                                                               tf_json)
                init_op = tf.initialize_all_variables()
                saver = tf.train.Saver()
            sv = tf.train.Supervisor(is_chief=(flag_task_index == 0),
                                     logdir=log_dir,
                                     init_op=init_op,
                                     summary_op=None,
                                     saver=saver,
                                     global_step=step_op,
                                     save_model_secs=3600)

            with sv.prepare_or_wait_for_session(server.target) as sess:
                # tf.get_default_graph().finalize()
                # 如果是同步模式
                if flag_task_index == 0 and issync == 1:
                    sv.start_queue_runners(sess, [chief_queue_runner])
                    sess.run(init_token_op)

                tic = time.time()
                log_cnt, total_loss = 1, 0.0
                reader = readers[flag_task_index]
                pylog.info(f"train files:{reader.filenames}")
                for features, labels in reader:
                    feed = trans_feat_feed(features, self.model.input_ph_dict)
                    _, step, loss = sess.run([train_op, step_op, loss_op], feed_dict=feed)
                    total_loss += loss
                    if step % self.steps_log == 0:
                        toc = time.time()
                        log_info = []
                        batch_cost = round((toc - tic) * 1000 / self.steps_log, 2)
                        row_cost = round((toc - tic) * 1000000 / (self.steps_log * self.batch_size), 2)
                        log_info.extend([f"row:{log_cnt * self.batch_size // 10000}w"])
                        log_info.extend([f"steps:{step // 1000}k", f"log_cnt:{log_cnt}"])
                        log_info.extend([f"loss:{np.round(total_loss / log_cnt, 5)}"])
                        log_info.extend([f"{batch_cost}:ms/batch", f"{row_cost}:μs/row"])
                        pylog.info(", ".join(log_info))
                        log_cnt += 1
                        total_loss = 0.0
                        tic = time.time()
                pylog.info(f"train files finish....")
                # sv.stop()
